/****************************************************************************
 * Copyright (C) 2010 Laboratoire de Recherche Informatique (LRI)
 * All Rights Reserved
 *
 *     This source code is distributed under the terms of the
 *     GNU General Public License.
 ****************************************************************************/

/*
 * main.c
 * 2010/08/24 by Feng LOU & Peter CLOTE
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <limits.h>
#include "../lib/misc.h"
#include "../lib/struct.h"
#include "../lib/rnabor.h"
#include "../lib/pf.h"

/*
 * read dot.ps from RNAfold
 * got the base pair probability P and nucleotide unpaired probability Q
 */
PQ* getDotPlot(char* rna) {
	int n = strlen(rna);
	double** P = (double**) malloc(n * sizeof(double*));
	double *Q = (double*) malloc(n * sizeof(double));
	int i, j;
	for (i = 0; i < n; i++) {
		P[i] = (double*) malloc(n * sizeof(double));
		for (j = 0; j < n; j++) {
			P[i][j] = 0;
		}
		Q[i] = 0;
	}
	// run external program RNAfold, and produce dot.ps
	char cmd[10240];
	FILE *file;
	sprintf(cmd, "echo %s | RNAfold -p ", rna);
	system(cmd);
	file = fopen("dot.ps", "r");
	char c[1000] = "";
	int indice_word;
	int is_sequence = 0;
	int is_ubox = 0;
	double p;
	char** words;
	char* w;
	if (file != NULL) {
		while (fgets(c, 1000, file) != NULL) {
			// split each line "c" into char** word
			words = (char**) malloc(4 * sizeof(char*));
			int indice_w = 0;
			w = strtok(c, " ");
			while (w != NULL) {
				if (indice_w < 4) {
					words[indice_w] = w;
				}
				indice_w++;
				w = strtok(NULL, " ");
			}
			// check RNA sequence
			if (is_sequence == 1) {
				char* sequence = words[0];
				// eliminate "\" at the end of line
				int size = strlen(rna);
				sequence[size] = '\0';
				//assert(strcmp(sequence, rna) == 0);
				assert(size == n);
				is_sequence = 0;
			}
			// get base pairs probability P
			if (indice_w == 4) {
				if (strncmp(words[3], "ubox", 4) == 0) {
					i = atoi(words[0]);
					j = atoi(words[1]);
					p = pow(atof((words[2])), 2);
					P[i - 1][j - 1] = p;
				}
			}
			if (strcmp(words[0], "/sequence") == 0) {
				is_sequence = 1;
			}
			free(words);
			free(w);
		}
		char cmd[256];
		sprintf(cmd, "rm *.ps");
		system(cmd);
		fclose(file);
	}
	// get nucleotide unpaired probability Q
	double sum;
	for (i = 0; i < n; i++) {
		sum = 0.0;
		for (j = 0; j < n; j++) {
			if (i < j) {
				sum += P[i][j];
			} else if (j < i) {
				sum += P[j][i];
			}
		}
		Q[i] = 1 - sum;
	}
	PQ * pq = create_PQ(P, Q, n);
	return pq;
}

/*------------------------------------------------------------
 * The following 3 functions, case1(), case2() and case3() used
 * in function computeEnergyMatrix(rna,S0,Kmax,M,P,Q,alpha,beta)
 *------------------------------------------------------------*/
int case1(list *L0, int i, int j) {
	int num = 0;
	item* parser = L0->first;
	while (parser != NULL) {
		if (parser->x >= i && parser->y == j)
			num += 1;
		parser = parser->next;
	}
	if (DEBUG) {
		list *X = init_list();
		list *Y = init_list();
		parser = L0->first;
		while (parser != NULL) {
			if (i <= parser->x && parser->y <= j - 1)
				add_list_item(X, parser->x, parser->y);
			if (i <= parser->x && parser->y <= j)
				add_list_item(Y, parser->x, parser->y);
			parser = parser->next;
		}
		int diff = 0;
		parser = X->first;
		while (parser != NULL) {
			if (existItemInList(Y, parser->x, parser->y) == 0)
				diff++;
			parser = parser->next;
		}
		parser = Y->first;
		while (parser != NULL) {
			if (existItemInList(X, parser->x, parser->y) == 0)
				diff++;
			parser = parser->next;
		}
		printf("case1, %d, %d, %d, %d", i, j, num, diff);
		assert(num == diff);
	}
	return num;
}
int case2(list *L0, int i, int j) {
	int num = 0;
	if (existItemInList(L0, i, j) == 0)
		num += 1;
	item* parser = L0->first;
	while (parser != NULL) {
		if (parser->x == i && parser->y < j)
			num += 1;
		if (i < parser->x && parser->y == j)
			num += 1;
		parser = parser->next;
	}
	if (DEBUG) {
		list *X = init_list();
		list *Y = init_list();
		add_list_item(X, i, j);
		parser = L0->first;
		while (parser != NULL) {
			if (i + 1 <= parser->x && parser->y <= j - 1)
				add_list_item(X, parser->x, parser->y);
			if (i <= parser->x && parser->y <= j)
				add_list_item(Y, parser->x, parser->y);
			parser = parser->next;
		}
		int diff = 0;
		parser = X->first;
		while (parser != NULL) {
			if (existItemInList(Y, parser->x, parser->y) == 0)
				diff++;
			parser = parser->next;
		}
		parser = Y->first;
		while (parser != NULL) {
			if (existItemInList(X, parser->x, parser->y) == 0)
				diff++;
			parser = parser->next;
		}
		printf("case2, %d, %d, %d, %d", i, j, num, diff);
		assert(num == diff);
	}
	return num;
}
int case3(list *L0, int i, int j, int r) {
	assert((i < r) && (r <= j - THRESHOLD - 1));
	int num = 0;
	if (existItemInList(L0, r, j) == 0)
		num += 1;
	item* parser = L0->first;
	while (parser != NULL) {
		if (i <= parser->x && parser->x <= r - 1 && r + 1 <= parser->y
				&& parser->y <= j - 1)
			num += 1;
		if (parser->x == r && r + 1 <= parser->y && parser->y <= j - 1)
			num += 1;
		if (i <= parser->x && parser->x != r && (parser->y == r || parser->y
				== j))
			num += 1;
		parser = parser->next;
	}
	if (DEBUG) {
		list *X = init_list();
		list *Y = init_list();
		add_list_item(X, r, j);
		parser = L0->first;
		while (parser != NULL) {
			if (i <= parser->x && parser->y <= r - 1)
				add_list_item(X, parser->x, parser->y);
			if (r + 1 <= parser->x && parser->y <= j - 1)
				add_list_item(X, parser->x, parser->y);
			if (i <= parser->x && parser->y <= j)
				add_list_item(Y, parser->x, parser->y);
			parser = parser->next;
		}
		int diff = 0;
		parser = X->first;
		while (parser != NULL) {
			if (existItemInList(Y, parser->x, parser->y) == 0)
				diff++;
			parser = parser->next;
		}
		parser = Y->first;
		while (parser != NULL) {
			if (existItemInList(X, parser->x, parser->y) == 0)
				diff++;
			parser = parser->next;
		}
		printf("case3, %d, %d, %d, %d, %d", i, j, r, num, diff);
		assert(num == diff);
	}
	return num;
}
/*
 * Usage information of Program
 */
void usage(void) {
	fprintf(stderr, "Usage:\n\n"
		"RNAborMEA -f filename -o [0|1|2]\n\n"
		"-f: \t file consists of optional FASTA comment, RNA, sec str, sec str constraints\n"
		"each on separate line. Additionally, weights alpha,beta,Kmax can\n"
		"be given on a last line, separated by tabs.\n" 
                "\nA string containing constraints on the structure encoded with the symbols:\n"
                " * :    (the base is unpaired) and matching brackets\n" 
                "( ):    (base i pairs base j)\n"
                " . :    (no constraint for this base).\n"
                "\nFor example:\n"
		">EMBL1\n"
		"GGGGCCCCC\n"
		"(.......)\n"
                "(...**..)\n"
		"2  1  10\n"
		"-- Optionally, the last line can not be given, in which case,\n"
		"alpha=1=beta and Kmax=len(rna). For example:\n"
		">EMBL1\n"
		"GGGGCCCCC\n"
		"(.......)\n"
                "(...**..)\n"
		"-- Optionally, in the last line, only alpha and beta can be given,\n"
		"in which case Kmax = len(rna)\n"
		"-- RNA is upper or lowercase sequence of A,C,G,U (or T)\n"
		"0 <= alpha,beta are weights of P[i][j] resp Q[i]\n"
		"default values are alpha = 2, beta = 1\n"
		"\n-o: \t 1 calculate only MEA(k) for all distances\n"
		" \t 2 calculate only the partition function\n"
		" \t 0 calculate both MEA(k) and the partition function\n"
		" \t default value is 0\n\n");
}
/*
 * main function
 */
void run(char *rna, char *secStr, char *id, char *constraint, int option,
		int alpha, int beta, int Kmax) {
	PQ* pq = getDotPlot(rna);
	if (option == 0 || option == 1)
		computeMEA_RNAbor(rna, secStr, Kmax, pq->P, pq->Q, alpha, beta, id,
				constraint);
	if (option == 0 || option == 2)
		computeBoltzmannMEAProb(rna, secStr, Kmax, (double) (strlen(rna)),
				pq->P, pq->Q, alpha, beta, id, constraint);
	free_PQ(pq);
}
void gotSeq(char *file, int option) {
	data * D = readInput(file);
	int i;
	for (i = 0; i < D->nb_seq; i++) {
		printf(
				"\n\nRNA sequence is:  %s with  %d  nucleotide \nStructure S0 is:  %s \n",
				D->rna[i], strlen(D->rna[i]), D->secStr[i]);
		printf("constraint is:  %s\n\n\n", D->constraint[i]);
		run(D->rna[i], D->secStr[i], D->id[i], D->constraint[i], option,
				D->alpha[i], D->beta[i], D->Kmax[i]);
	}
	/* free */
	free_data(D);
}
int main(int argc, char *argv[]) {
	if (argc < 3) {
		usage();
		exit(1);
	}
	// read command options
	int i;
	char *file = (char*) malloc(128 * sizeof(char));
	int option = 0;
	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-')
			switch (argv[i][1]) {
			case 'f':
				file = argv[i + 1];
				break;
			case 'o':
				option = atoi(argv[i + 1]);
				break;
			default:
				usage();
				exit(0);
			}
	}
	gotSeq(file, option);
	return 0;
}
