#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "../lib/misc.h"
#include "../lib/struct.h"

void error(char *s) {
	fprintf(stderr, "%s\n", s);
	exit(1);
}
/*
 * functions check if the 2 letter can be base pair
 */
int guPair(char x, char y) {
	return (x == 'G' && y == 'U' || x == 'U' && y == 'G');
}
int auPair(char x, char y) {
	return (x == 'A' && y == 'U' || x == 'U' && y == 'A');
}
int cgPair(char x, char y) {
	return (x == 'C' && y == 'G' || x == 'G' && y == 'C');
}
int a(int i, int j, char *s) {
	if (i < strlen(s) && j < strlen(s)) {
		if (auPair(s[i], s[j]))
			return AUbond;
		else if (cgPair(s[i], s[j]))
			return CGbond;
		else if (guPair(s[i], s[j]))
			return GUbond;
		else
			return INF;
	} else
		error("Indices out of bounds for RNA sequence\n");
}
int basePair(char x, char y) {
	return (auPair(x, y) || cgPair(x, y) || guPair(x, y));
}
int watsonCrick(char x, char y) {
	return (auPair(x, y) || cgPair(x, y));
}
/*
 * Get function parameters from file
 */
data * readInput(char *file) {
	// get sequence number
	int num_seq = 0;
	FILE * f = fopen(file, "r");
	char c;
	if (f == NULL) {
		printf("Cannot open file.\n");
		exit(1);
	}
	while ((c = fgetc(f)) != EOF) {
		if (c == '>') {
			num_seq += 1;
		}
	}
	fclose(f);
	data * d = init_data(num_seq);
	// get input
	int seq_index = -1;
	int line_index = 1;
	int c_index = 0;
	int parameter_number;
	char *p = (char *) malloc(16 * sizeof(char));
	f = fopen(file, "r");
	while ((c = fgetc(f)) != EOF) {
		if (c == '>') {
			seq_index += 1;
			line_index = 1;
			c_index = 0;
			parameter_number = 1;
		} else if (c == '\n') {
			if (line_index == 1) {
				// end of id line
				d->id[seq_index][c_index] = '\0';
			} else if (line_index == 2) {
				// end of sequence s0 line
				d->rna[seq_index][c_index] = '\0';
				d->Kmax[seq_index] = c_index;
			} else if (line_index == 3) {
				// end of structure line
				d->secStr[seq_index][c_index] = '\0';
			} else if (line_index == 4) {
				// end of constraint line
				d->constraint[seq_index][c_index] = '\0';
			} else if (line_index == 5 && parameter_number == 3) {
				p[c_index] = '\0';
				d->Kmax[seq_index] = atoi(p);
				parameter_number++;
			}
			c_index = 0;
			line_index += 1;
		} else if (line_index == 1) {
			// id line
			if (c != ' ') {
				d->id[seq_index][c_index] = c;
				c_index += 1;
			}
		} else if (line_index == 2) {
			// rna sequence line
			if (c == 't' || c == 'T') {
				c == 'U';
			}
			if (c == 'G' || c == 'C' || c == 'A' || c == 'U') {
				d->rna[seq_index][c_index] = toupper(c);
				c_index += 1;
			}
		} else if (line_index == 3) {
			// rna structure line
			if (c == '(' || c == '.' || c == ')') {
				d->secStr[seq_index][c_index] = c;
				c_index++;
			}
		} else if (line_index == 4) {
			// rna constraint line
			if (c == '(' || c == '.' || c == ')'|| c == '*') {
				d->constraint[seq_index][c_index] = c;
				c_index++;
			}
		} else if (line_index == 4) {
			// parameters line: alpha, beta, kmax
			if (c == ' ') {
				/* get alpha*/
				if (parameter_number == 1) {
					p[c_index] = '\0';
					d->alpha[seq_index] = atoi(p);
				} /* get beta*/
				else if (parameter_number == 2) {
					p[c_index] = '\0';
					d->beta[seq_index] = atoi(p);
				}
				/* get kmax*/
				else if (parameter_number == 3) {
					p[c_index] = '\0';
					d->Kmax[seq_index] = atoi(p);
				}
				parameter_number++;
				c_index = 0;
			} else {
				p[c_index] = c;
				c_index += 1;
			}
		}
	}
	fclose(f);
	free(p);
	return d;
}
/*
 * return a list of base pair found in secStr
 */
list * getBasePairList(char *secStr) {
	char c1;
	int c2, i;
	int n = strlen(secStr);
	list* L = init_list();
	int* S = (int*) malloc(n * sizeof(int));
	int size = 0;
	// Parser secStr
	for (i = 0; i < n; i++) {
		c1 = secStr[i];
		if (c1 == '(') {
			S[size] = i;
			size++;
		} else if (c1 == ')') {
			if (size == 0) {
				//return -1;
			} else {
				c2 = S[size - 1];
				add_list_item(L, c2, i);
				size--;
				L->size++;
			}
		}
	}
	free(S);
	if (size != 0) {/*return 1;*/
	}
	return L;
}
int basePiarDisntance(char* A, char* B) {
	list *listA = getBasePairList(A);
	list *listB = getBasePairList(B);
	int sum = 0;
	if (listA->size > 0) {
		item* parser = listA->first;
		while (parser != NULL) {
			if (existItemInList(listB, parser->x, parser->y) == 0)
				sum++;
			parser = parser->next;
		}
	}
	if (listB->size > 0) {
		item* parser = listB->first;
		while (parser != NULL) {
			if (existItemInList(listA, parser->x, parser->y) == 0)
				sum++;
			parser = parser->next;
		}
	}
	free_list(listA);
	free_list(listB);
	return sum;
}
int numBasePairs(char* secStr) {
	int count = 0;
	int size = strlen(secStr);
	int i;
	for (i = 0; i < size; i++) {
		if (secStr[i] == '(')
			count++;
	}
	return count;
}
/*
 *  Whether secStr is compatible with rna sequence
 *  secStr is string
 */
int compatible(char* rna, char* secStr) {
	if (strlen(rna) != strlen(secStr))
		return 0;
	list* l = getBasePairList(secStr);
	int a, b;
	if (l->size > 0) {
		item* parser = l->first;
		while (parser != NULL) {
			a = parser->x;
			b = parser->y;
			if (basePair(secStr[a], secStr[b]) == 0)
				return 0;
			parser = parser->next;
		}
		return 1;
	}
}
double evalViennaEnergy(char *rna, char *str) {
	// run RNAeval with rna and str
	char *tmpFileName = "tmp";
	FILE *tmpFile = fopen(tmpFileName, "w");
	fprintf(tmpFile, "%s\n", rna);
	fprintf(tmpFile, "%s\n", str);
	fclose(tmpFile);
	char cmd[10240];
	sprintf(cmd, "RNAeval < %s > tmpOF", tmpFileName);
	system(cmd);
	FILE *tmpOutFile = fopen("tmpOF", "r");
	char c;
	char e[256];
	int is_energy = 0;
	int e_indice = 0;
	int line_number = 1;
	while ((c = fgetc(tmpOutFile)) != EOF) {
		if (is_energy == 2 && c != ' ') {
			e[e_indice] = c;
			e_indice++;
		}
		if (c == ' ' && is_energy != 2) {
			is_energy = 1;

		}
		if (c == '(' && is_energy == 1) {
			is_energy = 2;
		}
		if (c == ')' && is_energy == 2) {
			is_energy = 0;
			e[e_indice - 1] = '\0';
		}
	}
	fclose(tmpOutFile);
	double energy = atof(e);
	sprintf(cmd, "rm tmp | rm tmpOF");
	system(cmd);
	return energy;
}
int validConstraint(char *c){

}
