#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../lib/misc.h"
#include "../lib/struct.h"

void initializePartitionFunctionMatrix(char *rna, char *S0, int Kmax,
		int alpha, int beta, double RT, double **P, double *Q, double ***Z) {
	list* secStrList0 = getBasePairList(S0);
	int n = strlen(rna);
	int i, j, k, d, r;
	double produit;
	int num;
	// now set to correct values
	for (d = 0; d < n; d++) { // d = 0, ... , n-1
		for (i = 0; i < n - d; i++) { // i = 0, ... , n-d-1
			j = i + d; // j = 0, ... , n-1
			if (j <= i + THRESHOLD) { // i<=j<=i+THRESHOLD
				produit = 1;
				for (r = i; r <= j; r++) {
					produit *= exp(beta * Q[r] / RT);
				}
				Z[i][j][0] = produit;
			} else if (j = i + THRESHOLD + 1) {
				produit = 1;
				num = 0;
				for (r = i + 1; r < j; r++) {
					produit *= exp(beta * Q[r] / RT);
				}
				if (existItemInList(secStrList0, i, j) == 1) {
					Z[i][j][0] = produit * exp(2 * alpha * P[i][j] / RT);
					Z[i][j][1] = produit * exp(beta * (Q[i] + Q[j]));
					Z[j][i][0] = 1.0;
					Z[j][i][1] = 1.0;
				} else {
					Z[i][j][0] = produit * exp(beta * (Q[i] + Q[j]) / RT);
					Z[j][i][0] = 1.0;
					if (basePair(rna[i], rna[j])) {
						Z[i][j][1] = produit * exp(2 * alpha * P[i][j] / RT);
						Z[j][i][1] = 1.0;
					}
				}
			}
		}
	}
	free_list(secStrList0);
}
void *** computePartitionFunctionMatrix(char *rna, char *S0, int Kmax,
		double RT, double ***Z, double **P, double *Q, int alpha, int beta) {
	int n = strlen(rna);
	double Zsum = 0.0;
	double Nsum = 0.0;
	double sum;
	int num, i, j, r, k, d, b0, b1, b2, k0, k1;
	list* secStrList0 = getBasePairList(S0);
	for (d = THRESHOLD + 2; d < n; d++) {
		for (i = 0; i < n - d; i++) {
			for (k = 0; k <= Kmax; k++) {
				j = i + d;
				sum = 0;
				num = 0;
				// Set Z[(i,j,k)] to sum of following cases
				// Case 1: j unpaired in S[i,j], where S in NOT S0, but arbitrary str
				b0 = case1(secStrList0, i, j);
				if (k - b0 >= 0) {
					sum += Z[i][j - 1][k - b0] * exp(beta * Q[j] / RT);
					num += Z[j - 1][i][k - b0];
				}
				//Case 2: (i,j) in S, where S is NOT S0, but arbitrary str
				if (basePair(rna[i], rna[j])) { // check if i,j can pair
					b1 = case2(secStrList0, i, j);
					//b1 = bpDist(S0[i+1,j-1] +{(i,j)},S0[i,j])
					if (k - b1 >= 0) {
						sum += Z[i + 1][j - 1][k - b1] * exp(2 * alpha
								* P[i][j] / RT);
						num += Z[j - 1][i + 1][k - b1];
					}
				}
				// Case 3: (r,j) in S, some i<r<j, where S is NOT S0, but arbitrary str
				for (r = i + 1; r < j - THRESHOLD; r++) {
					if (basePair(rna[r], rna[j])) {
						b2 = case3(secStrList0, i, j, r);
						// b2 = bpDist(S0[i,r-1]+S0[r+1,j-1]+{(r,j)},S0[i,j])
						for (k0 = 0; k0 < k - b2 + 1; k0++) {
							k1 = k - b2 - k0; // k0+k1+b2 = k
							if (k1 >= 0) {
								sum += Z[i][r - 1][k0] * Z[r + 1][j - 1][k1]
										* exp(2 * alpha * P[r][j] / RT);
								num += Z[r - 1][i][k0] * Z[j - 1][r + 1][k1];
							}
						}
					}
				}
				Z[i][j][k] = sum;
				Z[j][i][k] = num;
			}
		}
	}
	free_list(secStrList0);
}
void computeBoltzmannMEAProb(char *rna, char *S0, int Kmax, double RT,
		double **P, double *Q, int alpha, int beta, char *id, char *constraint) {
	/* ----- initialization  ----*/
	int n = strlen(rna);
	int i, j, k;
	double *BoltzProb = (double *) malloc((Kmax + 1) * sizeof(double));
	double *UnifProb = (double *) malloc((Kmax + 1) * sizeof(double));
	double ***Z = (double ***) malloc(n * sizeof(double **));
	for (i = 0; i < n; i++) {
		Z[i] = (double **) malloc(n * sizeof(double *));
		for (j = 0; j < n; j++) {
			Z[i][j] = (double *) malloc((Kmax + 1) * sizeof(double));
			for (k = 0; k <= Kmax; k++) {
				if (k == 0) {
					Z[i][j][k] = 1;
				} else {
					Z[i][j][k] = 0;
				}

			}
		}
	}
	initializePartitionFunctionMatrix(rna, S0, Kmax, alpha, beta, RT, P, Q, Z);
	computePartitionFunctionMatrix(rna, S0, Kmax, RT, Z, P, Q, alpha, beta);
	double Z_sum = 0;
	double N_sum = 0;
	for (k = 0; k <= Kmax; k++) {
		if (Z[0][n - 1][k] > 1) {
			Z_sum += Z[0][n - 1][k];
			N_sum += Z[n - 1][0][k];
		}
	}
	for (k = 0; k <= Kmax; k++) {
		if (Z[0][n - 1][k] > 1) {
			BoltzProb[k] = Z[0][n - 1][k] / Z_sum;
			UnifProb[k] = Z[n - 1][0][k] / N_sum;
		}
	}
	/* Write info into out.pf*/
	char filename[256];
	sprintf(filename, "out/%s.pf", id);
	FILE *fo = fopen(filename, "w");
	fprintf(fo, "Distance \t Z \t N \n");
	for (k = 0; k <= Kmax; k++) {
		if (Z[0][n - 1][k] > 1) {
			fprintf(fo, "%d \t %f \t %f\n", k, BoltzProb[k], UnifProb[k]);
		}
	}
	fclose(fo);
	free(BoltzProb);
	free(UnifProb);
	for (i = 0; i < n; i++) {
		for (j = 0; j < n; j++) {
			free(Z[i][j]);
		}
		free(Z[i]);
	}
	free(Z);
}
