/*
 * struct.c
 * structure type definitions and functions
 * 2010/08/24 by Feng LOU & Peter CLOTE
 */
#include <stdio.h>
#include <stdlib.h>
#include "../lib/struct.h"

/*
 * return a new char* copy from original
 */
char * copyTo_char(char *o) {
	int size = strlen(o);
	char* r = (char*) malloc((size + 1) * sizeof(char));
	int i;
	for (i = 0; i < size; i++) {
		r[i] = o[i];
	}
	r[size] = '\0';
	return r;
}
/*
 * list functions **
 * create a new list
 */
list * init_list() {
	list* l = (list*) malloc(sizeof(list));
	l->first = NULL;
	l->size = 0;
	return l;
}
/*
 * return the x value at position 'pos' of list
 */
int get_list_x(list *l, int pos) {
	if (pos >= l->size) {
		printf("bad position, try another smaller position!");
		return -1;
	} else {
		item* tmp = l->first;
		int i = 0;
		while (i != pos) {
			tmp = tmp->next;
			i++;
		}
		return tmp->x;
	}
}
/*
 * return the y value at position 'pos' of list
 */
int get_list_y(list *l, int pos) {
	if (pos >= l->size) {
		printf("bad position, try another smaller position!");
		return -1;
	} else {
		item* tmp = l->first;
		int i = 0;
		while (i != pos) {
			tmp = tmp->next;
			i++;
		}
		return tmp->y;
	}
}
/*
 * set the x value at position 'pos' of list
 */
int set_list_x(list *l, int pos, int x) {
	if (pos >= l->size) {
		printf("bad position, try another smaller position!");
		return -1;
	} else {
		item* tmp = l->first;
		int i = 0;
		while (i != pos) {
			tmp = tmp->next;
			i++;
		}
		tmp->x = x;
	}
	return 1;
}
/*
 * set the y value at position 'pos' of list
 */
int set_list_y(list *l, int pos, int y) {
	if (pos >= l->size) {
		printf("bad position, try another smaller position!");
		return -1;
	} else {
		item* tmp = l->first;
		int i = 0;
		while (i != pos) {
			tmp = tmp->next;
			i++;
		}
		tmp->y = y;
	}
	return 1;
}
/*
 * add a new item at the end of list
 */
void add_list_item(list *l, int x, int y) {
	item* a = (struct item*) malloc(sizeof(struct item));
	if (x > y) {
		int tmp = x;
		x = y;
		y = tmp;
	}
	(a)->x = x;
	(a)->y = y;
	(a)->next = NULL;
	if (l->size > 0) {
		item *tmp = l->first;
		while (tmp->next != NULL) {
			tmp = tmp->next;
		}
		tmp->next = a;
	} else
		l->first = a;
	l->size++;
}
/*
 * remove the item at position 'pos' of list
 */
void remove_list_item(list* l, int pos) {
	if (l->size == 0) {
		printf("empty list, no item to remove. \n");
		exit(0);
	} else if (pos > l->size - 1) {
		printf("bad position, try another position smaller than %d.\n", pos);
		exit(0);
	} else if (pos == 0) {
		if (l->size == 1) {
			free(l->first);
			l->first = NULL;
			l->size = 0;
		} else {
			item* tmp = l->first->next;
			free(l->first);
			l->first = tmp;
			l->size--;
		}
	} else {
		item* tmp = l->first;
		int i = 0;
		while (i != pos - 1) {
			tmp = tmp->next;
			i++;
		}
		item* next_item = tmp->next->next;
		free(tmp->next);
		tmp->next = next_item;
		l->size--;
	}
}
/*
 *  create a new list copy from original list
 */
list* copy_list(list* l) {
	list* tmp = init_list();
	item* parser = l->first;
	while (parser != NULL) {
		add_list_item(tmp, parser->x, parser->y);
		parser = parser->next;
	}
	return tmp;
}
/*
 *  if the item (i, j) exist in l return 1
 *  else return 0
 */
int existItemInList(list* l, int x, int y) {
	int r = 0;
	item* parser = l->first;
	while (parser != NULL) {
		if (parser->x == x && parser->y == y) {
			r = 1;
			return r;
		}
		parser = parser->next;
	}
	return r;
}
/*
 * print the x and y value of each item in list
 */
void print_list(list* l) {
	if (l->first == NULL) {
		printf("[]\n");
	} else {
		item* tmp = l->first;
		printf("[ ");
		while (tmp != NULL) {
			printf("(%d %d) ", tmp->x, tmp->y);
			tmp = tmp->next;
		}
		printf("]\n");
	}
}
/*
 * return all the base pair (x, y) from a Secondary Structure
 */
list* set_list(char* secStr) {
	list* tmp = init_list();
	int size = strlen(secStr);
	int k = 0;
	int s = 0;
	int left[size];
	// parser the secStr
	int i;
	for (i = 0; i < size; i++) {
		if (secStr[i] == '(') {
			left[k] = i;
			k = k + 1;
		} else if (secStr[i] == ')') {
			add_list_item(tmp, left[k - 1], i);
			k = k - 1;
			s = s + 1;
		}
	}
	tmp->size = s;
	return tmp;
}
/*
 * free memory used by list
 */
void free_list(list* l) {
	if (l->size == 0) {
		free(l);
	} else {
		item* parser = l->first;
		item* tmp;
		while (parser != NULL) {
			tmp = parser->next;
			l->first = tmp;
			free(parser);
			parser = tmp;
		}
		free(l);
	}
}
PQ * create_PQ(double **P, double *Q, int n) {
	PQ *pq = (PQ *) malloc(sizeof(PQ));
	pq->P = (double **) malloc(n * sizeof(double *));
	pq->Q = (double *) malloc(n * sizeof(double));
	int i, j;
	for (i = 0; i < n; i++) {
		pq->P[i] = (double *) malloc(n * sizeof(double));
		pq->Q[i] = Q[i];
		for (j = 0; j < n; j++) {
			pq->P[i][j] = P[i][j];
		}
	}
	pq->size = n;
	for (i = 0; i < n; i++) {
		free(P[i]);
	}
	free(P);
	free(Q);
	return pq;
}
void free_PQ(PQ *pq) {
	int i;
	for (i = 0; i < pq->size; i++) {
		free(pq->P[i]);
	}
	free(pq->P);
	free(pq->Q);
	free(pq);
}
void printDouble1D(double *T, int size) {
	int i;
	for (i = 0; i < size; i++) {
		printf("%d \t %f\n", i, T[i]);
	}
}
void printDouble2D(double **T, int size) {
	int i, j;
	for (i = 0; i < size; i++) {
		for (j = 0; j < size; j++) {
			printf("%d \t %d \t %f\n", i, j, T[i][j]);
		}
	}
}
data * init_data(int n) {
	printf("number of sequence is %d\n", n);
	data *D = (data *) malloc(sizeof(data));
	D->nb_seq = n;
	D->id = (char **) malloc(n * sizeof(char *));
	D->rna = (char **) malloc(n * sizeof(char *));
	D->secStr = (char **) malloc(n * sizeof(char *));
	D->constraint = (char **) malloc(n * sizeof(char *));
	D->alpha = (int *) malloc(n * sizeof(int));
	D->beta = (int *) malloc(n * sizeof(int));
	D->Kmax = (int *) malloc(n * sizeof(int));
	int i;
	for (i = 0; i < n; i++) {
		D->rna[i] = (char *) malloc(10240 * sizeof(char));
		D->secStr[i] = (char *) malloc(10240 * sizeof(char));
		D->id[i] = (char *) malloc(10240 * sizeof(char));
		D->constraint[i] = (char *) malloc(10240 * sizeof(char));
		D->alpha[i] = 1;
		D->beta[i] = 1;
		D->Kmax[i] = 0;
	}
	return D;
}
void free_data(data *d) {
	int i;
	for (i = 0; i < d->nb_seq; i++) {
		free(d->id[i]);
		free(d->rna[i]);
		free(d->secStr[i]);
		free(d->constraint[i]);
	}
	free(d->alpha);
	free(d->beta);
	free(d->Kmax);
	free(d->id);
	free(d->rna);
	free(d->secStr);
	free(d->constraint);
	free(d);
}
